/*
gcc julia-spiral-center.c `PKG_CONFIG_PATH=${HOME}/opt/lib/pkgconfig pkg-config --cflags --libs mandelbrot-numerics` -lm -fopenmp -O3 -std=c99 -Wall -Wextra -pedantic
LD_LIBRARY_PATH=${HOME}/opt/lib/ ./a.out |
ffmpeg -framerate 25 -i - -pix_fmt yuv420p -profile:v high -level:v 4.1 -crf:v 20 -movflags +faststart julia-spiral-center.mp4 -y
*/

#include <complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef M_PI
#define M_PI 3.141592653589793
#endif

// git clone https://code.mathr.co.uk/mandelbrot-numerics.git
// make -C mandelbrot-numerics/c/lib prefix=${HOME}/opt install
#include <mandelbrot-numerics.h>

double cnorm(double _Complex z)
{
  double x = creal(z);
  double y = cimag(z);
  return x * x + y * y;
}

int main(int argc, char **argv)
{
  // silence -Wunused-parameter
  (void) argc;
  (void) argv;
  // number of animation frames
  const int frames = 25 * 30 * 1;
  // parameters for Julia set graphics rendering
  const int iterations = 1024;
  const double escape_radius_squared = 1 << 24;
  const int width = 640;
  const int height = 360;
  unsigned char *pgm = malloc(width * height);
  // how far from the target component to place the moving c
  // 1.0 is the boundary of the component
  const double radius = 1.25;
  // period of the target component
  const int period = 3;
  // these don't need to be exact, they will be changed by m_d_interior()
  double _Complex z = 0;
  double _Complex c = -1.754;
  // maximum number of steps of Newton's method
  const int steps = 64;
  // needs a good initial guess to converge to the correct attractor
  double _Complex w = -0.1 - 0.01 * I;
  for (int frame = 0; frame < frames; ++frame)
  {
    // path around the exterior of the component
    double t = (frame + 0.5) / frames;
    double _Complex q = radius * cexp(2 * M_PI * I * t);
    // find c,z such that f_c^p(z) = z and d/dz f_c^p(z) = q
    m_d_interior(&z, &c, z, c, q, period, steps);
    // find w such that f_c^p(w) = w
    // use previous frame's w as initial guess for continuity
    m_d_attractor(&w, w, c, period, steps);
    // scaling for view
    double r = cabs(w);
    // render Julia set
    #pragma omp parallel for
    for (int j = 0; j < height; ++j)
    {
      for (int i = 0; i < width; ++i)
      {
        // view centered on w
        double _Complex z0 =
          ( ((i + 0.5) / width - 0.5) * 2 * width / height
            - I * ((j + 0.5) / height - 0.5) * 2 ) * r + w;
        // initialize derivative with pixel spacing
        double _Complex dz = 2 * r / height;
        // unescaped pixels will be grey
        pgm[j * width + i] = 128;
        for (int k = 0; k < iterations; ++k)
        {
          dz = 2 * z0 * dz;
          z0 = z0 * z0 + c;
          if (cnorm(z0) > escape_radius_squared)
          {
            // compute exterior distance estimate
            double de = 2 * cabs(z0) * log(cabs(z0)) / cabs(dz);
            // colour boundary based on distance estimate
            pgm[j * width + i] = 255 * tanh(de);
            break;
          }
        }
      }
    }
    // output PGM stream to stdout
    fprintf(stdout, "P5\n%d %d\n255\n", width, height);
    fwrite(pgm, width * height, 1, stdout);
    fflush(stdout);
  }
  // cleanup
  free(pgm);
  return 0;
}
